#include <stdio.h>
#include <unistd.h>
#include <netdb.h>
#include <ctype.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <signal.h>
#include <errno.h>
#include <wait.h>

#define IPA(x)	(((ntohl(x))&0xff000000)>>24)
#define IPB(x)	(((ntohl(x))&0xff0000)>>16)
#define IPC(x)	(((ntohl(x))&0xff00)>>8)
#define IPD(x)	((ntohl(x))&0xff)

#define T_STDIO	0
#define T_PROG	1
#define T_NET	2

#define SERVICES "/etc/services"

#ifdef SOLARIS
#define bzero(dest,len)		memset((dest),'\0',(len))
#define bcopy(src,dest,len)	memmove((dest),(src),(len))
#endif

int cflg=0;		/* -c option used */
int lflg=0;		/* -l option used */
int pflg=0;		/* -p option used */
int vflg=0;		/* -v option used */

int fcnt;		/* number of file descriptors */
int *fin;		/* file descriptor in */
int *fout;		/* file descriptor out */
int *ftype;		/* file descriptor type */
int *lfd;		/* listen() file descriptors */
char **host;		/* hosts */
int *port;		/* ports */
int *stfd;		/* stat of I/O */

int pid;		/* pid of child */
char line[BUFSIZ];	/* I/O buffer - also used for fatal messages */
char *argv0;		/* sanitized program name */
char *version="1.3";	/* version */

/*************************************************************************
* History of various programs by Bryan Dongray <btd@dongrays.com>
* 
* 1989
*    send	to open TCP socket to a remote host and send stdin
*    recv	to listen on TCP port and received bytes put to stdout
*	Used together to transfer file over raw network socket
*
* 1992
*    socket	to open to a remote host and send/recv to stdin/stdout
*		Good for inetd on port 25, replacing sendmail
*    sockthru	to listen on port, and forward to remote host:port
*		Standalone single process for port forwarding
*    listen	to listen and run a command to/from - like a mini inetd
*
* 1998
*    2net	v1.0	to listen or connect two TCP network ports
*			and transfer data between both
*		v1.1	added stdin/stdout ability (replaces "socket")
*			add -c for webproxies that exit if input closed
* 2002
*		v1.2	added connection to program possible
*			replaces "listen"
*		v1.3	allow more than two connections, renamed "thru"
*			for logging, net-splitting, etc
*************************************************************************/

char *strchr(), *strrchr();
void waitchild(), talkremote(), prusage();
int openarg();

main(argc, argv)
char **argv;
{
	char *p;
	int doagain;
	int fi;
	int ftot;

	/*
	 * Get a nice looking program name
	 */
	argv0=argv[0];
	if (p=strrchr(argv0, '/'))
		argv0 = &p[1];

	/*
	 * Hint: opts are done with lazy processing - all must be lonesome
	 * eg use "-c -l -v" cannot use "-clv"
	 */
	while (argc > 1) {

		/*
		 * -v	verbose
		 */
		if (strcmp(argv[1], "-v") == 0) {
			vflg++;
			argv++;
			argc--;
			continue;
		}

		/*
		 * -l	looping
		 */
		if (strcmp(argv[1], "-l") == 0) {
			lflg=1;
			argv++;
			argc--;
			continue;
		}

		/*
		 * -p	parallel looping
		 */
		if (strcmp(argv[1], "-p") == 0) {
			pflg=1;
			argv++;
			argc--;
			continue;
		}

		/*
		 * -c	close only once all is done
		 */
		if (strcmp(argv[1], "-c") == 0) {
			cflg=1;
			argv++;
			argc--;
			continue;
		}

		/*
		 * -r	restrict which connection communicates to others
		 *	-rAB means input from A does not go to B - useful for logging
		 */
		if (strncmp(argv[1], "-r", 2) == 0) {
/* Hmmmmmmmmm
thinking how to do this */
			argv++;
			argc--;
			continue;
		}

		/*
		 * unknown arg
		 */
		if (argv[1][0] == '-' && argv[1][1] != '\0') {
			fprintf(stderr, "%s: unknown arg: %s\n", argv0, argv[1]);
			prusage();
		}

		/*
		 * no more args
		 */
		break;
	}

	/*
	 * show version - if requested
	 */
	if (vflg)
		fprintf(stderr, "%s: version %s\n", argv0, version);

	/*
	 * Must have last least one arg
	 */
	if ((fcnt=argc-1) == 0) {
		fprintf(stderr, "%s: no connections\n", argv0);
		prusage();
	}

	/*
	 * -l and -p make no sense together
	 */
	if (lflg && pflg) {
		fprintf(stderr, "%s: Cannot have -p and -l\n", argv0);
		prusage();
	}

	/*
	 * Create arrays for args
	 */
	if ((fin=(int *)calloc(fcnt, sizeof (int))) == NULL) {
		fprintf(stderr, "%s: calloc() error\n", argv0);
		exit(1);
	}
	if ((fout=(int *)calloc(fcnt, sizeof (int))) == NULL) {
		fprintf(stderr, "%s: calloc() error\n", argv0);
		exit(1);
	}
	if ((ftype=(int *)calloc(fcnt, sizeof (int))) == NULL) {
		fprintf(stderr, "%s: calloc() error\n", argv0);
		exit(1);
	}
	if ((lfd=(int *)calloc(fcnt, sizeof (int))) == NULL) {
		fprintf(stderr, "%s: calloc() error\n", argv0);
		exit(1);
	}
	if ((host=(char **)calloc(fcnt, sizeof (char *))) == NULL) {
		fprintf(stderr, "%s: calloc() error\n", argv0);
		exit(1);
	}
	if ((port=(int *)calloc(fcnt, sizeof (int))) == NULL) {
		fprintf(stderr, "%s: calloc() error\n", argv0);
		exit(1);
	}
	if ((stfd=(int *)calloc(fcnt, sizeof (int))) == NULL) {
		fprintf(stderr, "%s: calloc() error\n", argv0);
		exit(1);
	}

	/*
	 * Parse args into host and ports
	 */
	ftot=0;
	for (fi=0; fi < fcnt; fi++)
		ftot += parsearg('A'+fi, argv[fi+1], &host[fi], &port[fi]);
	if (ftot)
		prusage();

	/*
	 * Count stdin/stdout connection types
	 */
	ftot=0;
	for (fi=0; fi < fcnt; fi++)
		if (host[fi] == NULL)
			ftot++;
	if (ftot > 1) {
		fprintf(stderr, "%s: Cannot have more than one stdin/stdout\n", argv0);
		prusage();
	}

	/*
	 * If any programs, we need to wait on the children
	 */
	pid = -1;
	signal(SIGCHLD, waitchild);

	doagain=1;
	while (doagain) {
		/*
		 * open connections
		 */
		for (fi=0; fi < fcnt; fi++) {
			if (vflg)
				fprintf(stderr, "%s: Opening %c: ", argv0, 'A'+fi);
			ftype[fi]=openarg(fi, host[fi], port[fi], &fin[fi], &fout[fi]);
		}

		/*
		 * note if we are going to loop
		 */
		doagain=(lflg || pflg);

		/*
		 * if parallel, fork the child
		 */
		if (pflg) {
			switch (pid=fork()) {
			case -1:
				/*
				 * oh dear, fork failed: fall back on -l method
				 */
				sprintf(line, "%s: fork", argv0);
				perror(line);
				fprintf(stderr, "%s: will retry after this close\n", argv0);
				break;

			case 0:
				/*
				 * child does the I/O (and will not loop)
				 */
				doagain=0;
				break;

			default:
				/*
				 * parent to get next one ready
				 */
				if (vflg)
					fprintf(stderr, "%s: fork pid %d\n", argv0, pid);
				/*
				 * cannot have childs file descriptors open
				 */
				for (fi=0; fi < fcnt; fi++) {
					if (ftype[fi] != T_STDIO) {
						if (fin[fi] > 1)
							close(fin[fi]);
						if (fout[fi] > 1)
							close(fout[fi]);
					}
				}

				/*
				 * back to the top of the loop
				 */
				continue;
			}
		}

		/*
		 * if we are not looping, close any listener
		 */
		if (!doagain)
			for (fi=0; fi < fcnt; fi++)
				if (lfd[fi])
					close(lfd[fi]);

		/*
		 * do the chit-chat
		 */
		talkremote();

		/*
		 * that's all folks
		 */
		if (vflg)
			if (pid)
				fprintf(stderr, "%s: closed\n", argv0);
			else
				fprintf(stderr, "%s: %d closed\n", argv0, getpid());
	}
	exit(0);
}

/*
 * waitchild() - get child exit status and pid
 */
void
waitchild()
{
	int wpid, wstat;

	wpid=wait(&wstat);
	if (vflg)
		fprintf(stderr, "%s: pid %d exit %d\n", argv0, wpid, WEXITSTATUS(wstat));
	signal(SIGCHLD, waitchild);
}

/*
 * parsearg() - work out hostname and port
 */
int
parsearg(whch, arg, hostp, portp)
char whch, *arg, **hostp;
int *portp;
{
	char *p, *q, *s;
	static FILE *fpserv=NULL;

	if (strcmp(arg, "-") == 0) {
		/*
		 * "-" means stdin/stdout
		 */
		*hostp=NULL;
		*portp=0;
		if (vflg)
			fprintf(stderr, "%s: Connection %c: stdin/stdout\n", argv0, whch);
	} else if (p=strchr(*hostp=arg, ':')) {
		/*
		 * host:port is connect to port on remote host
		 * :port is listen on local port
		 */
		q=p;
		p++;

		/*
		 * Cannot have two colons
		 */
		if (strchr(p, ':')) {
			fprintf(stderr, "%s: Connection %c: format error \"%s\" - more than one colon\n", argv0, whch, arg);
			return 1;
		}

		/*
		 * Need port information
		 */
		if (*p == '\0') {
			fprintf(stderr, "%s: Connection %c: format error \"%s\" - no port\n", argv0, whch, arg);
			return 1;
		}

		/*
		 * Analyze port
		 */
		*portp=0;
		while (*p) {
			if (!isdigit(*p)) {
				/*
				 * port is more than just a number
				 * look in SERVICES file
				 */
				if (fpserv) {
					rewind(fpserv);
				} else {
					fpserv=fopen(SERVICES, "r");
				}
				if (fpserv) {
					while (fgets(line, sizeof (line), fpserv)) {
						/* eg
						 telnet	23/tcp	comment
						 */

						/*
						 * remove comments
						 */
						if (p=strchr(line, '#'))
							*p='\0';
						/*
						 * skip to find spaces
						 */
						for (p=line; *p && !isspace(*p); p++)
							;
						/*
						 * if end of line, next!
						 */
						if (!*p)
							continue;
						/*
						 * first space is end of name
						 */
						*p++ = '\0';
						/*
						 * is it a match?
						 */
						if (strcmp(line, &q[1]) != 0)
							continue;
						/*
						 * skip spaces
						 */
						while (isspace(*p))
							p++;
						/*
						 * no digit, no port
						 */
						if (!isdigit(*p))
							continue;
						s=p;
						while (isdigit(*p))
							p++;
						/*
						 * after digits needs /tcp
						 */
						if (strncmp(p, "/tcp", 4) != 0)
							continue;
						/*
						 * we have a port number
						 */
						*portp=atoi(s);
						break;
					}
				}
				/*
				 * if never found a port number
				 */
				if (*portp == 0) {
					fprintf(stderr, "%s: Connection %c: %s - port unknown\n", argv0, whch, arg);
					return 1;
				}
				break;
			}
			p++;
		}
		*q++ = '\0';
		/*
		 * convert sequence of numbers to port number
		 */
		if (*portp == 0)
			*portp=atoi(q);
		if (vflg)
			if (**hostp == '\0')
				fprintf(stderr, "%s: Connection %c: listen on port %d\n", argv0, whch, *portp);
			else
				fprintf(stderr, "%s: Connection %c: remote to %s on port %d\n", argv0, whch, *hostp, *portp);
	} else {
		/*
		 * otherwise it is a command
		 */
		*hostp=arg;
		*portp=0;
		if (vflg)
			fprintf(stderr, "%s: Connection %c: executable: %s\n", argv0, whch, arg);
	}
	return 0;
}

/*
 * openarg() - open connection
 */
int
openarg(ind, hoststr, portno, pfi, pfo)
int ind;
char *hoststr;
int portno;
int *pfi, *pfo;
{
	int s;
	struct sockaddr_in sin;
	int fromlen;
	struct sockaddr_in from;
	unsigned char ipstr[4];
	struct hostent *hp;
	int addr;
	int rwa[2], rwb[2];
	int wpid;

	if (hoststr == NULL) {
		/* stdin/out */
		*pfi=0;
		*pfo=1;
		if (vflg)
			fprintf(stderr, "stdin/stdout ok\n");
		return T_STDIO;
	}

	if (portno == 0) {
		/* executable */

		/*
		 * two pipes
		 */
		if (pipe(rwa) == -1 || pipe(rwb) == -1) {
			if (vflg)
				sprintf(line, "pipe");
			else
				sprintf(line, "%s: pipe", argv0);
			perror(line);
			exit(1);
		}
		switch(wpid=fork()) {
		case -1:
			/*
			 * fork failed - no good
			 */
			if (vflg)
				sprintf(line, "fork");
			else
				sprintf(line, "%s: fork", argv0);
			perror(line);
			exit(1);
		case 0:
			/*
			 * child needs to set stdin/stdout with the pipes
			 */
			close(0);
			close(1);
			dup2(rwa[0], 0);
			dup2(rwb[1], 1);
			/*
			 * close everything else
			 */
			for (s=3; s < 256; s++)
				close(s);
			/*
			 * run the program
			 */
			execlp("/bin/sh", "sh", "-c", hoststr, 0);
			if (vflg)
				sprintf(line, "exec");
			else
				sprintf(line, "%s: exec", argv0);
			perror(line);
			exit(1);
		}
		if (vflg)
			fprintf(stderr, "pid=%d: %s\n", wpid, hoststr);
		/*
		 * parent sets its end of the pipes
		 */
		close(rwa[0]);
		close(rwb[1]);
		*pfi=rwb[0];
		*pfo=rwa[1];
		return T_PROG;
	}

	/*
	 * see if we have been listening before
	 */
	if (lfd[ind]) {
		s=lfd[ind];
		goto nextlisten;
	}

	/*
	 * all network connections need a socket
	 */
	bzero((char *) &sin, sizeof(sin));
	sin.sin_family=AF_INET;
	sin.sin_port=htons(portno);
	if ((s=socket(AF_INET, SOCK_STREAM, 0)) == -1) {
		if (vflg)
			sprintf(line, "%s:%d socket", hoststr, portno);
		else
			sprintf(line, "%s: %s:%d socket", argv0, hoststr, portno);
		perror(line);
		exit(1);
	}

	if (*hoststr == '\0') {
		/*
		 * listening type
		 */
		sin.sin_addr.s_addr=INADDR_ANY;
		if (bind(s, (struct sockaddr *)&sin, sizeof(sin)) == -1) {
			if (vflg)
				sprintf(line, "%s:%d bind", hoststr, portno);
			else
				sprintf(line, "%s: %s:%d bind", argv0, hoststr, portno);
			perror(line);
			exit(1);
		}
		if (listen(s, SOMAXCONN) == -1) {
			if (vflg)
				sprintf(line, "%s:%d listen", hoststr, portno);
			else
				sprintf(line, "%s: %s:%d listen", argv0, hoststr, portno);
			perror(line);
			exit(1);
		}
		lfd[ind]=s;
nextlisten:
		if (vflg)
			fprintf(stderr, "listening on port %d\n", portno);
redolisten:
		/*
		 * accept the connection
		 */
		fromlen=sizeof(from);
		if ((*pfi=accept(s, &from, &fromlen)) == -1) {
			if (errno == EINTR)
				goto redolisten;
			sprintf(line, "%s: %s:%d accept", argv0, hoststr, portno);
			perror(line);
			exit(1);
		}
		*pfo = *pfi;
		if (vflg) {
			ipstr[0]=IPA(from.sin_addr.s_addr);
			ipstr[1]=IPB(from.sin_addr.s_addr);
			ipstr[2]=IPC(from.sin_addr.s_addr);
			ipstr[3]=IPD(from.sin_addr.s_addr);
			if (hp=gethostbyaddr(ipstr, 4, AF_INET))
				fprintf(stderr, "%s: connect from %s (%u.%u.%u.%u) port %d\n", argv0, hp->h_name, ipstr[0], ipstr[1], ipstr[2], ipstr[3], from.sin_port);
			else
				fprintf(stderr, "%s: connect from %u.%u.%u.%u port %d\n", argv0, ipstr[0], ipstr[1], ipstr[2], ipstr[3], from.sin_port);
		}
		return T_NET;
	}

	/*
	 * connecting to remote
	 */
	if ((addr=inet_addr(hoststr)) == -1 || addr == 0) {
		/*
		 * lookup a hostname
		 */
		if ((hp=gethostbyname(hoststr)) == NULL) {
			if (vflg)
				sprintf(line, "%s:%d host not found", hoststr, portno);
			else
				sprintf(line, "%s: %s:%d host not found", argv0, hoststr, portno);
			fprintf(stderr, "%s\n", line);
			exit(1);
		}
		bcopy(hp->h_addr, (char *)&sin.sin_addr, hp->h_length);
	} else
		/*
		 * use the IP
		 */
		sin.sin_addr.s_addr=addr;

	/*
	 * Make the connection
	 */
	if (connect(s, &sin, sizeof(sin)) == -1) {
		if (vflg)
			sprintf(line, "%s:%d connect", hoststr, portno);
		else
			sprintf(line, "%s: %s:%d connect", argv0, hoststr, portno);
		perror(line);
		exit(1);
	}
	*pfi=s;
	*pfo=s;
	if (vflg)
		fprintf(stderr, "%s: %s:%d connected\n", argv0, hoststr, portno);
	return T_NET;
}

/*
 * talkremote() - two way read/write
 */
void
talkremote()
{
	fd_set readmask;
	int len;
	int fi, fo;

	/*
	 * set file descriptors as sctive
	 */
	for (fi=0; fi < fcnt; fi++)
		stfd[fi]=1;

	for (;;) {
		/*
		 * set readmask
		 */
		FD_ZERO(&readmask);
		len=0;
		for (fi=0; fi < fcnt; fi++)
			if (stfd[fi]) {
				len++;
				FD_SET(fin[fi], &readmask);
			}
		/*
		 * if nothing left
		 */
		if (len == 0)
			break;

		/*
		 * wait for any I/O
		 */
		if (select(sizeof (int)*8, &readmask, 0, 0, 0) == -1) {
			sprintf(line, "%s: select", argv0);
			perror(line);
			break;
		}

		/*
		 * whichever input
		 */
		for (fi=0; fi < fcnt; fi++)
			if (stfd[fi] && FD_ISSET(fin[fi], &readmask)) {
				/*
				 * read it
				 */
				switch (len=read(fin[fi], line, sizeof (line))) {
				case -1:
				case 0:
shutit:
					/*
					 * a close - close in style needed
					 */
					if (vflg)
						if (pid)
							fprintf(stderr, "%s: %c closed\n", argv0, fi+'A');
						else
							fprintf(stderr, "%s: %d %c closed\n", argv0, getpid(), fi+'A');
					stfd[fi]=0;
					if (!cflg) {
						if (ftype[fi] == T_PROG)
							close(fin[fi]);
						else if (ftype[fi] == T_NET)
							shutdown(fin[fi], 0);
						/*
						 * pass on close
						 */
						for (fo=0; fo < fcnt; fo++)
							if (fo != fi)
								if (ftype[fo] == T_PROG)
									close(fin[fo]);
								else if (ftype[fo] == T_NET)
									shutdown(fin[fo], 1);
					}
					break;
				default:
					/*
					 * write to others
					 */
					for (fo=0; fo < fcnt; fo++)
						if (fo != fi)
							if (write(fout[fo], line, len) == -1)
								goto shutit;
					break;
				}
				if (vflg > 1)
					if (pid)
						fprintf(stderr, "%s: read %c %d\n", argv0, fi+'A', len);
					else
						fprintf(stderr, "%s: %d read %c %d\n", argv0, getpid(), fi+'A', len);
			}
	}
	if (vflg > 1)
		if (pid)
			fprintf(stderr, "%s: read end\n", argv0);
		else
			fprintf(stderr, "%s: %d read end\n", argv0, getpid());

	/*
	 * force close on necessary file descriptors
	 */
	for (fi=0; fi < fcnt; fi++)
		if (ftype[fi] != T_STDIO) {
			close(fin[fi]);
			close(fout[fi]);
		}
	return;
}

/*
 * prusage() - print usage message and exit
 */
void
prusage()
{
	fprintf(stderr, "Usage: %s [-c] [-l] [-p] [-v] {[<host>]:<port>|-|PROG}...\n", argv0);
	fprintf(stderr, "\t-c\tdo not close connections until all are ready to close\n");
	fprintf(stderr, "\t\tuseful for connections to web proxies\n");
	fprintf(stderr, "\t-l\tloop after close (no listen until loop)\n");
	fprintf(stderr, "\t\tto keep listening again - but disabled during the loop\n");
	fprintf(stderr, "\t-p\tloop after open (listen stays open)\n");
	fprintf(stderr, "\t\tlike your own single port inetd listener\n");
	fprintf(stderr, "\t-v\tverbose\n");
	fprintf(stderr, "\t\t-v -v for very verbose\n");
	fprintf(stderr, "\tConnections:\n");
	fprintf(stderr, "\t\thost:port\tconnect to remote host on port\n");
	fprintf(stderr, "\t\t:port\t\tlisten for connection on port\n");
	fprintf(stderr, "\t\t-\t\tstdin/stdout\n");
	fprintf(stderr, "\t\tPROG\t\texecute with stdin/stdout connected\n");
	fprintf(stderr, "\thost\thostname or IP\n");
	fprintf(stderr, "\tport\tnumeric or TCP in %s\n", SERVICES);
	fprintf(stderr, "\tPROG\tbest to use quotes - called with \"sh -c\"\n");
	exit(2);
}
